import random
import typing

import numpy as np
import albumentations as A
import typing_extensions as tx

BboxParams = A.BboxParams(format="pascal_voc", label_fields=["categories"])

AugmentedResult = tx.TypedDict(
    "AugmentedResult",
    {
        "image": np.ndarray,
        "bboxes": typing.List[typing.Tuple[int, int, int, int]],
        "categories": typing.List[str],
    },
)


# pylint: disable=too-few-public-methods
class AugmenterProtocol(tx.Protocol):
    """A protocol defining how we expect augmentation
    pipelines to behave. bboxes is expected to be in
    pascal_voc (or x1, y1, x2, y2) format."""

    def __call__(
        self,
        image: np.ndarray,
        bboxes: typing.List[typing.Tuple[int, int, int, int]],
        categories: typing.List[str],
    ) -> AugmentedResult:
        pass


class RandomCropToPaddedBBox(A.DualTransform):
    """Crop an image to have a bounding box centered within it. Useful for
    cases where labels are trusted in the area directly near a bounding box
    but not far from them.
    Args:
        width: The width of the cropped area.
        height: The height of the cropped area.
        p (float): probability of applying the transform. Default: 1.
    Targets:
        image, bboxes
    Image types:
        uint8, float32
    """

    def __init__(self, width, height, always_apply=False, p=1.0):
        super().__init__(always_apply, p)
        self.width = width
        self.height = height

    def apply(self, img, **params):
        x_min, y_min = [params[k] for k in ["x_min", "y_min"]]
        img_h, img_w = img.shape[:2]
        x_min_i = max(int(min(x_min * img_w, img_w - self.width)), 0)
        y_min_i = max(int(min(y_min * img_h, img_h - self.height)), 0)
        x_max_i = x_min_i + self.width
        y_max_i = y_min_i + self.height
        crop = img[y_min_i:y_max_i, x_min_i:x_max_i]
        if x_max_i <= img_w and y_max_i <= img_h:
            return crop
        return np.pad(
            crop, ((0, max(y_max_i - img_h, 0)), (0, max(x_max_i - img_w, 0)), (0, 0))
        )

    def get_params_dependent_on_targets(self, params):
        img_h, img_w = params["image"].shape[:2]
        box_h, box_w = min(1, self.height / img_h), min(1, self.width / img_w)
        if (
            len(params["bboxes"]) == 0
        ):  # less likely, this class is for use with bboxes.
            # Choose a random region
            x_min = int(random.random() * (1 - box_w))
            y_min = int(random.random() * (1 - box_h))
            return {
                "x_min": x_min,
                "y_min": y_min,
            }
        x1b, y1b, x2b, y2b = params["bboxes"][
            round(random.random() * (len(params["bboxes"]) - 1))
        ][:4]
        xcb, ycb = (x1b + x2b) / 2, (y1b + y2b) / 2
        x_min = min(max(0, xcb - box_w / 2), 1 - box_w)
        y_min = min(max(0, ycb - box_h / 2), 1 - box_h)
        return {
            "x_min": x_min,
            "y_min": y_min,
        }

    @property
    def targets_as_params(self):
        return ["image", "bboxes"]

    def apply_to_bbox(self, bbox, **params):
        x_min, y_min, cols, rows = [
            params[k] for k in ["x_min", "y_min", "cols", "rows"]
        ]
        crop_w = self.width / cols
        crop_h = self.height / rows
        scale_x = 1 / crop_w
        scale_y = 1 / crop_h
        x1, y1, x2, y2 = [
            min(max(value - offset, 0), max_dim) * scale
            for value, offset, max_dim, scale in zip(
                bbox,
                [x_min, y_min, x_min, y_min],
                [crop_w, crop_h, crop_w, crop_h],
                [scale_x, scale_y, scale_x, scale_y],
            )
        ]
        return (x1, y1, x2, y2)

    def get_transform_init_args_names(self):
        return ("width", "height")

    def apply_to_keypoint(self, *args, **kwargs):
        raise NotImplementedError("Keypoints are not supported.")



# type: ignore

import torch
import torchvision
import effdet
import omegaconf


class EfficientDet(torch.nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        config = omegaconf.OmegaConf.merge(
            effdet.get_efficientdet_config(model_name=MODEL_NAME),
            omegaconf.OmegaConf.create(
                {
                    "min_level": MIN_LEVEL,
                    "max_level": MAX_LEVEL,
                    "num_levels": NUM_LEVELS,
                    "max_detection_points": MAX_DETECTION_POINTS,
                }
            ),
        )
        self.model = effdet.create_model_from_config(
            config=config,
            num_classes=NUM_CLASSES,
            bench_task="",
            pretrained=False,
            pretrained_backbone=False,
            image_size=(INPUT_HEIGHT, INPUT_WIDTH),
            max_det_per_image=MAX_DET_PER_IMAGE,
        )
        self.config = {"anchors": effdet.anchors.Anchors.from_config(self.model.config)}
        self.resize = torchvision.transforms.Resize(size=(INPUT_HEIGHT, INPUT_WIDTH))

    def forward(self, x):
        resized = self.resize(x)
        scales = torch.tensor(
            [
                [i.shape[2] / o.shape[2], i.shape[1] / o.shape[1]]
                for i, o in zip(x, resized)
            ]
        ).tile((1, 2))
        y = self.model(resized)
        class_out, box_out = y
        class_out, box_out, indices, classes = effdet.bench._post_process(
            class_out,
            box_out,
            num_levels=self.model.config.num_levels,
            num_classes=self.model.config.num_classes,
            max_detection_points=self.model.config.max_detection_points,
        )
        img_scale, img_size = None, None
        detections = effdet.bench._batch_detection(
            class_out.shape[0],
            class_out.cpu(),
            box_out.cpu(),
            self.config["anchors"].boxes.cpu(),
            indices.cpu(),
            classes.cpu(),
            img_scale,
            img_size,
            max_det_per_image=self.model.config.max_det_per_image,
            soft_nms=True,
        )
        return [
            {
                "boxes": group[:, :4] * scaler,
                "labels": group[:, 5].type(torch.IntTensor),
                "scores": group[:, 4],
            }
            for group, scaler in zip(detections, scales)
        ]



# type: ignore

import torch
import torchvision


BACKBONE_TO_CONSTRUCTOR = {
    "resnet50": torchvision.models.detection.fasterrcnn_resnet50_fpn,
    "mobilenet_large": torchvision.models.detection.fasterrcnn_mobilenet_v3_large_fpn,
    "mobilenet_large_320": torchvision.models.detection.fasterrcnn_mobilenet_v3_large_320_fpn,
}


class FasterRCNNObjectDetector(torch.nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = BACKBONE_TO_CONSTRUCTOR[BACKBONE_NAME](
            pretrained=False, num_classes=NUM_CLASSES, pretrained_backbone=False
        )
        self.model.transform.fixed_size = (INPUT_HEIGHT, INPUT_WIDTH)  # type: ignore
        self.model.transform.min_size = (min(INPUT_WIDTH, INPUT_HEIGHT),)  # type: ignore
        self.model.transform.max_size = max(INPUT_HEIGHT, INPUT_WIDTH)  # type: ignore

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)



import torch
import torchvision
import numpy as np
import typing_extensions as tx
import pkg_resources

from .. import datasets as mds
from .. import core as mc
from . import detector


BACKBONE_TO_CONSTRUCTOR = {
    "resnet50": torchvision.models.detection.fasterrcnn_resnet50_fpn,
    "mobilenet_large": torchvision.models.detection.fasterrcnn_mobilenet_v3_large_fpn,
    "mobilenet_large_320": torchvision.models.detection.fasterrcnn_mobilenet_v3_large_320_fpn,
}


class FasterRCNN(detector.Detector):
    """A wrapper around the FasterRCNN models in torchvision."""

    def __init__(
        self,
        annotation_config=mds.COCOAnnotationConfiguration90,
        pretrained_backbone: bool = True,
        pretrained_top: bool = False,
        backbone: tx.Literal[
            "resnet50", "mobilenet_large", "mobilenet_large_320"
        ] = "resnet50",
        device="cpu",
        resize_method: detector.ResizeMethod = "fit",
    ):
        super().__init__(device=device, resize_method=resize_method)
        self.annotation_config = annotation_config
        self.model = BACKBONE_TO_CONSTRUCTOR[backbone](
            pretrained=pretrained_top,
            progress=True,
            num_classes=len(annotation_config) + 1,
            pretrained_backbone=pretrained_backbone,
        ).to(self.device)
        self.set_input_shape(
            width=min(self.model.transform.min_size),  # type: ignore
            height=min(self.model.transform.min_size),  # type: ignore
        )
        self.backbone_name = backbone

    @property
    def training_model(self):
        """Training model for this detector."""
        return self.model

    @property
    def serve_module_string(self):
        return (
            pkg_resources.resource_string(
                "mira", "detectors/assets/serve/fasterrcnn.py"
            )
            .decode("utf-8")
            .replace("NUM_CLASSES", str(len(self.annotation_config) + 1))
            .replace("INPUT_WIDTH", str(self.input_shape[1]))
            .replace("INPUT_HEIGHT", str(self.input_shape[0]))
            .replace("BACKBONE_NAME", f"'{self.backbone_name}'")
        )

    @property
    def serve_module_index(self):
        return {
            **{0: "__background__"},
            **{
                str(idx + 1): label.name
                for idx, label in enumerate(self.annotation_config)
            },
        }

    def compute_inputs(self, images):
        images = np.float32(images) / 255.0
        return (
            torch.tensor(images, dtype=torch.float32)
            .permute(0, 3, 1, 2)
            .to(self.device)
        )

    def invert_targets(self, y, threshold=0.5, **kwargs):
        return [
            [
                mc.Annotation(
                    category=self.annotation_config[int(labelIdx) - 1],
                    x1=x1,
                    y1=y1,
                    x2=x2,
                    y2=y2,
                    score=score,
                )
                for (x1, y1, x2, y2), labelIdx, score in zip(
                    labels["boxes"].detach().cpu().numpy(),
                    labels["labels"].detach().cpu().numpy(),
                    labels["scores"].detach().cpu().numpy(),
                )
                if score > threshold
            ]
            for labels in y
        ]

    def set_input_shape(self, width, height):
        self._input_shape = (height, width, 3)
        self.model.transform.fixed_size = (height, width)  # type: ignore
        self.model.transform.min_size = (min(width, height),)  # type: ignore
        self.model.transform.max_size = max(height, width)  # type: ignore

    @property
    def input_shape(self):
        return self._input_shape

    def compute_targets(self, annotation_groups):
        return [
            {
                "boxes": torch.tensor(b[:, :4], dtype=torch.float32).to(self.device),
                "labels": torch.tensor(b[:, -1] + 1, dtype=torch.int64).to(self.device),
            }
            for b in [
                self.annotation_config.bboxes_from_group(g) for g in annotation_groups
            ]
        ]
